#include <iostream>
#include <windows.h>
#include <scrnsave.h>
#include "system/exception.h"
#include "system/singletons.h"
#include "singletons.h"

using namespace std;

extern void InitBase();

//this might as well only exist for the sake of abstrating the different entry points ...
int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPSTR lpszCmdLine, int nCmdShow) {
	sysBindSingletons();

	try {

		//get args
		int nargs = 0;
		LPWSTR *argv = CommandLineToArgvW(GetCommandLineW(), &nargs);
		if (!argv) {
			throw Error() << "WinScreenSaverSystem::init failed to get command line";
		} else {
			for (int i = 0; i < nargs; i++) {
				size_t len = wcslen(argv[i])+1;
				char *str = new char[len+1];
				assert(str);
				memset(str, 0, len+1);
				for (int j = 0; j < len; j++) {
					str[j] = (char)argv[i][j];
				}
				winSsSys.args.push_back(str);
				delete str;
			}
			LocalFree(argv);
		}

		//interpret screensaver args
		//	(and remove them?)
		enum SSMethod { smConfig, smPassword, smPreview, smSaver };
		SSMethod method = smConfig;
		HWND methodWnd = 0;
		for (int i = 1; i < winSsSys.args.size(); i++) {
			if (!_stricmp(winSsSys.args[i].c_str(), "/c")) {
				method = smConfig;
				int hwndInt = 0;
				if (i < winSsSys.args.size()-1) {
					i++;
					hwndInt = atoi(winSsSys.args[i].c_str());
				}
				methodWnd = hwndInt ? (HWND)hwndInt : GetForegroundWindow();
			} else if (!_stricmp(winSsSys.args[i].c_str(), "/s")) {
				method = smSaver;
			} else if (!_stricmp(winSsSys.args[i].c_str(), "/p") || !_stricmp(winSsSys.args[i].c_str(), "/l")) {
				method = smPreview;
				int hwndInt = 0;
				if (i < winSsSys.args.size()-1) {
					i++;
					hwndInt = atoi(winSsSys.args[i].c_str());
				}
				methodWnd = (HWND)hwndInt;
			} else if (!_stricmp(winSsSys.args[i].c_str(), "/a")) {
				method = smPassword;
				int hwndInt = 0;
				if (i < winSsSys.args.size()-1) {
					i++;
					hwndInt = atoi(winSsSys.args[i].c_str());
				}
				methodWnd = (HWND)hwndInt;
			}
		}

		switch (method) {
		case smConfig:
			//do config.  more specifically, run it in a window.
			//		if it's glush then have some ifdefs and casting and, for ss in-ss have no menu
			//				for ss not-in-ss have the menu
			break;
		case smPassword:
			//change password
			{
				HINSTANCE hmpr = LoadLibraryA("MPR.DLL");
				if (!hmpr) throw Error() << "MPR.DLL not found: cannot change password.";
				typedef VOID (WINAPI *PWDCHANGEPASSWORD)(LPCSTR,HWND,UINT,UINT);
				PWDCHANGEPASSWORD PwdChangePassword = (PWDCHANGEPASSWORD)GetProcAddress(hmpr, "PwdChangePasswordA");
				if (!PwdChangePassword) {
					FreeLibrary(hmpr);
					throw Error() << "PwdChangeProc not found: cannot change password";
				}
				PwdChangePassword("SCRSAVE", methodWnd, 0, 0);
				FreeLibrary(hmpr);
				return 1;	//don't do main loop
			}
			break;
		case smPreview:	//for preview methodWnd will exist
			winSsSys.setParentWnd(methodWnd);
		case smSaver:	//for saver methodWnd won't exist
			winSsSys.setInScreenSaver(true);
			ShowCursor(FALSE);
			break;
		}

		UINT oldval = 0;
		if (method == smSaver) {
			SystemParametersInfo(SPI_SETSCREENSAVERRUNNING, 1, &oldval, 0);
		}

		InitBase();

		if (method == smSaver) {
			SystemParametersInfo(SPI_SETSCREENSAVERRUNNING,0,&oldval,0);
		}
	} catch (Exception &t) {
		cout << "caught an exception: " << t << endl;
		return 1;
	}

	ShowCursor(TRUE);

	return 0;
}

