#include <iostream>	//include this before stdio or it complains about exit being defined twice
#include <stdio.h>	//NULL
#include "system/memory.h"	//the system/ prefix means 'we gotta pull this from the generic-system code'
#include "system/singletons.h"
#include "system/exception.h"
#include "system.h"
#include "timer.h"
#include "singletons.h"
#include "wstring.h"

using namespace std;

//////////////////////////////// keyboard system mapping

//the header file had 127 or so virtual keys.  maybe 256 maybe 512
//someone emailed me about an out of bounds virtual key value: 15xx
//if Windows has any more keys than this, I might have to resort to using stl maps ...
static Key::Key_t keyForVirtKey[2048];

static void initKeyMaps() {

	for (int i = 0; i < numberof(keyForVirtKey); i++) {
		keyForVirtKey[i] = (Key::Key_t)-1;
	}

#define VKBRIDGE2(x,y)			keyForVirtKey[VK_##x] = Key::y;
#define VKBRIDGE(x)				keyForVirtKey[VK_##x] = Key::x; 
#define VKBRIDGERANGE(x,y)		for (int i = VK_##x; i <= VK_##y; i++) keyForVirtKey[i] = (Key::Key_t)(i - VK_##x + Key::x);
#define VKBRIDGERANGE2(x,y,z)	for (int i = VK_##x; i <= VK_##y; i++) keyForVirtKey[i] = (Key::Key_t)(i - VK_##x + Key::z);
	for (int i = 0; i < 127; i++) {
		if (!strlen(Key::name[i])) continue;	//if we're not mapping the key anyways then skip it, huh?
		int vk = VkKeyScan(i);
		if (vk > numberof(keyForVirtKey)) {
			//what this says is there exists virtual keycodes that windows could return which are larger than
			// our keyForVirtKey map above.
			//maybe I should just use a stl map rather than allocating this huge array?  that would save on space
			// but would slow down our lookup time from constant- to logarithm.
			//meh, it's only windows.  not like things don't run slow enough already.
			cerr << "required virtual key index " << vk << " for key " << i << "/" << (char)i << "." << endl
				<< "I need to up the maximum keyForVirtKey array to include this size" << endl;
			exit(0);
		}
		keyForVirtKey[vk] = (Key::Key_t)i;
	}
	VKBRIDGE2(DELETE,DEL)
	VKBRIDGERANGE(F1, F15)
	//windows wont do two buttons... i tihnk the l and r shfit keys are there so ppl can get frustrated like I did ...
	VKBRIDGE(LSHIFT) VKBRIDGE(RSHIFT)
	VKBRIDGE2(SHIFT,LSHIFT)
	VKBRIDGE2(LCONTROL,LCTRL) VKBRIDGE2(RCONTROL,RCTRL)
	VKBRIDGE2(CONTROL,LCTRL)
	VKBRIDGE2(LMENU,LALT)
	VKBRIDGE2(RMENU,RALT)
	VKBRIDGERANGE2(NUMPAD0,NUMPAD9,KP0)
	VKBRIDGE(UP)
	VKBRIDGE(DOWN)
	VKBRIDGE(LEFT)
	VKBRIDGE(RIGHT)
	VKBRIDGE2(CAPITAL,CAPSLOCK)
	VKBRIDGE2(LWIN, LMETA)
	VKBRIDGE2(RWIN, RMETA)
}

//////////////////////////////// class code

void WinScreenSaverSystem::reset() {
	//init data values here
	hwnd = NULL;
	hParentWnd = NULL;
	hDC = NULL;
	hRC = NULL;
	//devmode anyone?

	loopCallback = NULL;
	exitRequested = false;
	
	initSize.set(720,480);
	initPos.set(	//depends on initSize
		(GetSystemMetrics(SM_CXSCREEN) >> 1) - (initSize.x >> 1),
		(GetSystemMetrics(SM_CYSCREEN) >> 1) - (initSize.y >> 1));
	initBPP = 0;
	initFullScreen = false;
	inScreenSaver = false;
	initDBuf = true;
	initTitle = "Windows Window";
}

//dirty of me...
#ifndef WM_MOUSEWHEEL	//???
#define WM_MOUSEWHEEL                   0x020A
#endif

static LONG WINAPI MySsWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static bool gotFirstMouseMove = false;
	static vec2i lastMousePos;
	static PAINTSTRUCT ps;

	switch(uMsg) {

		//screensaver specific
	case WM_ACTIVATE:
	case WM_ACTIVATEAPP:
	case WM_NCACTIVATE:
		if (winSsSys.getInScreenSaver() && LOWORD(wParam) == WA_INACTIVE) {
			winSsSys.requestExit();
		}
		break;
	case WM_SYSCOMMAND:
		if (winSsSys.getInScreenSaver()) {
			if (wParam == SC_SCREENSAVE) return FALSE;
			if (wParam == SC_CLOSE) return FALSE;
		}
		break;

		//rest of the story (similar to winsystem)
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;

	case WM_SIZE:
		disp.resize(LOWORD(lParam),HIWORD(lParam));
		if (winSsSys.getInScreenSaver()) break;
		return 0;

	case WM_CLOSE:
		sys.requestExit();
		return 0;

		///all the same deal
	case WM_MOUSEWHEEL:
		mouse.setButtonPress(mouse.BUTTON_LEFT,		!!(wParam & MK_LBUTTON));
		mouse.setButtonPress(mouse.BUTTON_MIDDLE,	!!(wParam & MK_MBUTTON));
		mouse.setButtonPress(mouse.BUTTON_RIGHT,	!!(wParam & MK_RBUTTON));
		mouse.setCursorPos(LOWORD(lParam),HIWORD(lParam));
		mouse.accumDZ(HIWORD(wParam));	//scroll
		if (winSsSys.getInScreenSaver()) {
			winSsSys.requestExit();
			break;
		}
		return 0;

	case WM_MOUSEMOVE:
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
		{
			mouse.setButtonPress(mouse.BUTTON_LEFT,		!!(wParam & MK_LBUTTON));
			mouse.setButtonPress(mouse.BUTTON_MIDDLE,	!!(wParam & MK_MBUTTON));
			mouse.setButtonPress(mouse.BUTTON_RIGHT,	!!(wParam & MK_RBUTTON));
			vec2i newMousePos = vec2i(LOWORD(lParam), HIWORD(lParam));
			vec2i mouseDel = lastMousePos - newMousePos;
			mouse.setCursorPos(VEC2ELEM(newMousePos));
			lastMousePos = newMousePos;
			if (winSsSys.getInScreenSaver()) {
				if (!!(wParam & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON))) {
					winSsSys.requestExit();
					break;
				}
				if (uMsg == WM_MOUSEMOVE) {
					if (!gotFirstMouseMove) {
						gotFirstMouseMove = true;
					} else if (mouseDel.x || mouseDel.y) {
						winSsSys.requestExit();
					}
				}
			}
		}
		return 0;

// ********* KEYBOARD WINDOWS MESSAGE ********* 

	case WM_KEYDOWN:
		assert(wParam >= 0 && wParam < numberof(keyForVirtKey));
		kbd.keyDown(keyForVirtKey[wParam]);
		if (winSsSys.getInScreenSaver()) {
			winSsSys.requestExit();
			break;
		}
		return 0;

	case WM_KEYUP:
		assert(wParam >= 0 && wParam < numberof(keyForVirtKey));
		kbd.keyUp(keyForVirtKey[wParam]);
		if (winSsSys.getInScreenSaver()) {
			winSsSys.requestExit();
			break;
		}
		return 0;
	}
	return (LONG)DefWindowProc(hWnd, uMsg, wParam, lParam); 
} 

void WinScreenSaverSystem::init() {

	//store local parameters
	dbuf = initDBuf;

	memSystemInit();

	//init our timer whenever
	timer.init();

	initTime = timer.getTime();

// BEGIN WINDOWS INIT register our window

	try {
		wstring wndClassName = str2wstr("WinScreenSaverSystemWindowClass");

		HINSTANCE hInst = GetModuleHandle(NULL);
		if (!hInst) throw Exception() << "failed to get HINSTANCE";

		WNDCLASS wc;
		memset(&wc, 0, sizeof(wc));
		wc.style         = CS_OWNDC;
		wc.lpfnWndProc   = MySsWndProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInst;
		wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
		wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = wndClassName.c_str();

		if (!RegisterClass(&wc)) throw Exception() << "RegisterClass() failed";

		DWORD dwExStyle =	WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		DWORD dwStyle =		WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

		wstring initTitleW = str2wstr(initTitle);

		if (inScreenSaver) {
			initPos.set(0,0);
			initSize.set(
				GetSystemMetrics(SM_CXSCREEN),
				GetSystemMetrics(SM_CYSCREEN));
			dwStyle = WS_POPUP | WS_VISIBLE;
			dwExStyle = WS_EX_TOPMOST;
		}

		hwnd = CreateWindowEx(dwExStyle, wndClassName.c_str(), initTitleW.c_str(), dwStyle, VEC2ELEM(initPos), VEC2ELEM(initSize), hParentWnd, NULL, hInst, NULL);
		if (!hwnd) throw Exception() << "failed to create window";

		hDC = GetDC(hwnd);
		if (!hDC) throw Exception() << "GetDC failed";

		PIXELFORMATDESCRIPTOR pfd;
		memset(&pfd, 0, sizeof(pfd));
		int iPixelFormat = GetPixelFormat(hDC);			 		//get our current format
		DescribePixelFormat(hDC, iPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);				//modify it

		pfd.nSize		= sizeof(pfd);
		pfd.nVersion	= 1;
		pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
		if (dbuf) pfd.dwFlags |= PFD_DOUBLEBUFFER;
		pfd.iPixelType	= PFD_TYPE_RGBA;
		pfd.iLayerType	= PFD_MAIN_PLANE;
		//FIXME: stencil RW stutters like mad when run in 16 bit color - not HW accelerated? whats up?
		pfd.cColorBits = 32;
		pfd.cAlphaBits = 8;
		pfd.cStencilBits = 16;

		int pf = ChoosePixelFormat(hDC, &pfd);								//find the best fit pixel format to set it to
		if (!pf) throw Exception() << "ChoosePixelFormat failed";
		if (!SetPixelFormat(hDC, pf, &pfd)) throw Exception() << "SetPixelFormat failed";		//and set it accordingly

		hRC = wglCreateContext(hDC);												//make the GL RC
		if (!hRC) throw Exception() << "wglCreateContext failed";
		if (!wglMakeCurrent(hDC, hRC)) throw Exception() << "wglMakeCurrent failed";		//and make it the current RC

		//ResetViewport
		//set up the init viewport size
		RECT r;
		if (!GetClientRect(hwnd, &r)) throw Exception() << "GetClientRect failed";
		initSize.set(r.right, r.bottom);		//should we store this here or in a sep variable?

		//good to go - show the window and leave
		ShowWindow(hwnd, SW_SHOW);

		//initialize fullscreen as false - and save current DEVMODE info
		EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &devmodeBackup);
		initFullScreen = false;

	} catch (Exception &t) {
		throw Error() << "WinScreenSaverSystem::init failed! " << t;
	}

	//keyboard stuff
	initKeyMaps();
}

void WinScreenSaverSystem::shutdown() {
	//shutdown windows
	wglMakeCurrent(NULL, NULL);
	if (hwnd) ReleaseDC(hwnd, hDC);
	if (hRC) wglDeleteContext(hRC);

	//shutdown any objects *HERE*
	memSystemShutdown();

	//and reset...
	reset();
}

void WinScreenSaverSystem::setInitDoubleBuffered(bool db) { initDBuf = db; }
void WinScreenSaverSystem::setInitPos(const vec2i &_initPos) { initPos = _initPos; }
void WinScreenSaverSystem::setInitSize(const vec2i &_initSize) { initSize = _initSize; }
void WinScreenSaverSystem::setInScreenSaver(bool inScreenSaver_) { inScreenSaver = inScreenSaver_; }

const std::string &WinScreenSaverSystem::getArg(int index) const { return args[index]; }
size_t WinScreenSaverSystem::getArgCount() const { return args.size(); }

void WinScreenSaverSystem::requestExit() {
	exitRequested = true;
}

void WinScreenSaverSystem::loop(void (*lc)()) {
	MSG msg;
	//store the main loop's callback method
	loopCallback = lc;

	while (!exitRequested) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			winSsSys.doLoop();
		}
	}
}

void WinScreenSaverSystem::doLoop() {
	preUpdate();
	if (loopCallback) loopCallback();
	postUpdate();
}

void WinScreenSaverSystem::preUpdate() {
	//system update init
	disp.update();	//update resolution
	mouse.update();
	kbd.preUpdate();
}

void WinScreenSaverSystem::postUpdate() {
	//system update shutdown
	kbd.postUpdate();
	//NOTICE - does this apply to all, or only to the local viewport?
	updateDisplayBuffers();
}

void WinScreenSaverSystem::updateDisplayBuffers() {
	if (dbuf) {
		SwapBuffers(hDC);
	} else {
		glFlush();
	}
}
